#!/usr/bin/env python

# Read in a polyhedron description, and output Flash ActionScript
# code suitable for compiling with MTASC (i.e. a .as file) into a
# dedicated viewer application for that polyhedron.

import sys
import os
import string
import random
from math import pi, asin, atan2, cos, sin, sqrt
from crosspoint import crosspoint

args = sys.argv[1:]

if len(args) > 0:
    infile = open(args[0], "r")
    args = args[1:]
else:
    infile = sys.stdin

if len(args) > 0:
    outfile = open(args[0], "w")
    args = args[1:]
else:
    outfile = sys.stdout

vertices = {}
faces = {}
normals = {}

lineno = 0
while 1:
    s = infile.readline()
    if s == "": break
    sl = s.split()
    lineno = lineno + 1
    if sl[0] == "point" and len(sl) == 5:
        vertices[sl[1]] = (float(sl[2]), float(sl[3]), float(sl[4]))
    elif sl[0] == "face" and len(sl) == 3:
        if sl[2] not in vertices:
            sys.stderr.write("line {:d}: vertex {} not defined\n".format(
                lineno, sl[2]))
        else:
            if sl[1] not in faces:
                faces[sl[1]] = []
            faces[sl[1]].append(sl[2])
    elif sl[0] == "normal" and len(sl) == 5:
        if sl[1] not in faces:
            sys.stderr.write("line {:d}: face {} not defined\n".format(
                lineno, sl[1]))
        else:
            normals[sl[1]] = (float(sl[2]), float(sl[3]), float(sl[4]))
    else:
        sys.stderr.write("line {:d}: unrecognised line format\n".format(lineno))
        continue
infile.close()

def realprint(a):
    for i in range(len(a)):
        outfile.write(str(a[i]))
        if i < len(a)-1:
            outfile.write(" ")
        else:
            outfile.write("\n")

def asprint(*a):
    realprint(a)

# Normalise the radius of our polyhedron so that it just fits
# inside the unit sphere.
maxlen = 0
for v in vertices.values():
    vlen = sqrt(v[0]**2 + v[1]**2 + v[2]**2)
    if maxlen < vlen: maxlen = vlen
for v in vertices.keys():
    vv = vertices[v]
    vertices[v] = (vv[0]/maxlen, vv[1]/maxlen, vv[2]/maxlen)

# Assign integer indices to vertices and faces.
vindex = {}
findex = {}
vlist = []
flist = []
for v in vertices.keys():
    vindex[v] = len(vlist)
    vlist.append(v)
for f in faces.keys():
    findex[f] = len(flist)
    flist.append(f)

# We expect the template for the Flash source to be called
# flashpoly.as and to reside in the same directory as this script.
# Python helpfully lets us know that directory at all times.
template = os.path.join(sys.path[0], "flashpoly.as");
tempfile = open(template, "r")
while 1:
    line = tempfile.readline()
    if line == "": break
    if "--- BEGIN POLYHEDRON ---" in line:
        # Instead of the test polyhedron description in the
        # template file, substitute our own.
        for i in range(len(vertices)):
            v = vertices[vlist[i]]
            asprint("       var p{:d} = {{ x:{:.17g}, y:{:.17g},"
                    " z:{:.17g} }};".format(i, v[0], v[1], v[2]))
        for i in range(len(faces)):
            f = faces[flist[i]]
            n = normals[flist[i]]
            fstr = ",".join(["p{:d}".format(vindex[v]) for v in f])
            asprint("       var f{:d} = {{ points: new Array({}), x:{:.17g}, "
                    "y:{:.17g}, z:{:.17g} }};".format(
                        i, fstr, n[0], n[1], n[2]))
        vstr = ",".join(["p{:d}".format(i) for i in range(len(vertices))])
        fstr = ",".join(["f{:d}".format(i) for i in range(len(faces))])
        asprint("       poly = {{ points: new Array({}), "
                "faces: new Array({}) }};".format(vstr, fstr))
        while 1:
            line = tempfile.readline()
            if line == "": break
            if "--- END POLYHEDRON ---" in line: break
    else:
        outfile.write(line)

tempfile.close()
